@@ -4,7 +4,9 @@ use 5.010;
 
 package App::Music::ChordPro;
 
-our $VERSION = "0.72";
+use App::Music::ChordPro::Version;
+
+our $VERSION = $App::Music::ChordPro::Version::VERSION;
 
 =head1 NAME
 
@@ -49,6 +51,8 @@ package App::Music::ChordPro;
 
 sub ::run {
     my $options = app_setup( "ChordPro", $VERSION );
+    binmode(STDERR, ':utf8');
+    binmode(STDOUT, ':utf8');
     $options->{trace}   = 1 if $options->{debug};
     $options->{verbose} = 1 if $options->{trace};
     main($options);
@@ -59,7 +63,8 @@ sub main {
 
     # Establish backend.
     my $of = $options->{output};
-    if ( $of ) {
+
+    if ( defined($of) && $of ne "" ) {
         if ( $of =~ /\.pdf$/i ) {
             $options->{generate} ||= "PDF";
         }
@@ -83,6 +88,23 @@ sub main {
             $options->{generate} ||= "Debug";
         }
     }
+    elsif ( -t STDOUT ) {
+	# No output, and stdout is terminal.
+	# Derive output name from input name.
+	if ( @ARGV > 1 ) {
+	    # No default if more than one input document.
+	    die("Please use \"--output\" to specify the output file name\n");
+	}
+	my $f = $ARGV[0];
+	$f =~ s/\.\w+$/.pdf/;
+	$f .= ".pdf" if $f eq $ARGV[0];
+	$options->{output} = $f;
+	warn("Writing output to $f\n") if $options->{verbose};
+    }
+    else {
+	# Write output to stdout.
+	$options->{output} = "-";
+    }
 
     $options->{generate} ||= "PDF";
     my $pkg = "App::Music::ChordPro::Output::".$options->{generate};
@@ -110,7 +132,7 @@ sub main {
 	foreach my $c ( @{ App::Music::ChordPro::Chords::chordnames() } ) {
 	    if ( $c =~ /^(.[b#]?)/ and $1 ne $prev )  {
 		$prev = $1;
-		push( @body, { type => "chord-grids",
+		push( @body, { type => "diagrams",
 			       context => "",
 			       origin => "__CLI__",
 			       chords => [ @chords ]
@@ -120,7 +142,7 @@ sub main {
 	    push( @chords, $c );
 	}
 
-	push( @body, { type => "chord-grids",
+	push( @body, { type => "diagrams",
 		       context => "",
 		       origin => "__CLI__",
 		       chords => [ @chords ]
@@ -166,7 +188,39 @@ sub main {
 
 =item B<--about> (short: B<-A>)
 
-About ChordPro.
+Prints version information about the ChordPro program. No other
+processing will be done.
+
+=item B<--config=>I<JSON> (shorter: B<--cfg>)
+
+A JSON file that defines the behaviour of the program and the layout
+of the output. See L<App::Music::ChordPro::Config> for details.
+
+This option may be specified more than once. Each additional config
+file overrides the corresponding definitions that are currently
+active.
+
+=item B<--cover=>I<FILE>
+
+Prepends the contents of the named PDF document to the output. This can
+be used to produce documents with cover pages.
+
+=item B<--csv>
+
+When generating PDF output, also write a CSV file with titles and page
+numbers. Some tools, e.g., MobileSheets, can use the CSV to process
+the PDF as a collection of independent songs.
+
+The CSV has the same name  as the PDF, with extension C<.pdf> replaced
+by C<.csv>.
+
+=item B<--diagrams=>I<WHICH>
+
+Prints diagrams of chords used in a song.
+
+I<WHICH> can be C<all> to print all chords used, C<user> to only print
+the user-defined chords, and C<none> to suppress printing of chord
+diagrams.
 
 =item B<--encoding=>I<ENC>
 
@@ -180,6 +234,14 @@ Only prints lyrics. All chords are suppressed.
 Useful to make prints for singers and other musicians that do not
 require chords.
 
+=item B<--no-csv>
+
+Suppresses generating the CSV file. See B<--toc>.
+
+=item B<--no-toc>
+
+Suppresses the table of contents. See B<--toc>.
+
 =item B<--output=>I<FILE> (short: B<-o>)
 
 Designates the name of the output file where the results are written
@@ -194,10 +256,6 @@ correspond to one of the backends that are currently supported:
 
 Portable document format (PDF).
 
-If a table of contents is generated with the PDF, ChordPro also writes
-a CSV file containing titles and page numbers. This CSV file has the
-same name as the PDF, with extenstion C<pdf> replaced by <csv>.
-
 =item txt
 
 A textual representation of the input, mostly for visual inspection.
@@ -208,15 +266,6 @@ A functional equivalent version of the ChordPro input.
 
 =back
 
-=item B<--config=>I<JSON> (shorter: B<--cfg>)
-
-A JSON file that defines the behaviour of the program and the layout
-of the output. See L<App::Music::ChordPro::Config> for details.
-
-This option may be specified more than once. Each additional config
-file overrides the corresponding definitions that are currently
-active.
-
 =item B<--start-page-number=>I<N> (short: B<-p>)
 
 Sets the starting page number for the output.
@@ -228,20 +277,12 @@ Includes a table of contents.
 By default a table of contents is included in the PDF output when
 it contains more than one song.
 
-=item B<--no-toc>
-
-Suppresses the table of contents.
-
 =item B<--transpose=>I<N> (short: -x)
 
 Transposes all songs by I<N> semi-tones. Note that I<N> may be
 specified as B<+>I<N> to transpose upward, using sharps, or as
 B<->I<N> to transpose downward, using flats.
 
-=item B<--user-chord-grids>
-
-Prints chord grids of all user defined chords used in a song.
-
 =item B<--version> (short: B<-V>)
 
 Prints the program version and exits.
@@ -255,6 +296,9 @@ not all of them actually do something.
 
 Options marked with * are better specified in the config file.
 
+B<Note:> Chordii used the term _grid_ for chord diagrams. It
+should not be confused with ChordPro grids.
+
 =over 4
 
 =item B<--text-font=>I<FONT> (short: B<-T>) *
@@ -283,15 +327,15 @@ Sets the font size for the chord names.
 
 =item B<--chord-grid-size=>I<N> (short: B<-s>) *
 
-Sets chord grid size (the total width of a chord grid).
+Sets the total width of a chord diagram.
 
 =item B<--chord-grids>
 
-Prints chord grids of all chords used in a song.
+Prints chord diagrams of all chords used in a song.
 
 =item B<--no-chord-grids> (short: B<-G>) *
 
-Disables printing of chord grids of the chords used in a song.
+Disables printing of chord diagrams of the chords used in a song.
 
 =item B<--easy-chord-grids>
 
@@ -303,12 +347,16 @@ Not supported.
 
 =item B<--chord-grids-sorted> (short: B<-S>) *
 
-Prints chord grids of the chords used in a song, ordered by key and
+Prints chord diagrams of the chords used in a song, ordered by key and
 type.
 
 =item B<--no-chord-grids-sorted> *
 
-Prints chord grids in the order they appear in the song.
+Prints chord diagrams in the order they appear in the song.
+
+=item B<--user-chord-grids> *
+
+Prints chord diagrams of all user defined chords used in a song.
 
 =item B<--even-pages-number-left> (short B<-L>)
 
@@ -320,7 +368,7 @@ Prints even/odd pages with pages numbers left on odd pages.
 
 =item B<--page-size=>I<FMT> (short: B<-P>) *
 
-Specifies page size, e.g. C<a4> (default), C<letter>.
+Specifies the page size for the PDF output, e.g. C<a4> (default), C<letter>.
 
 =item B<--single-space> (short B<-a>)) *
 
@@ -419,8 +467,8 @@ Don't use the specific config file, even if it exists.
 
 Sets a configuration item. I<item> must be in the format of
 colon-separated configuration keys, an equal sign, and the value. For
-example, the equivalent of B<--no-chord-grids> is
-B<--define=chordgrid:show=0>.
+example, the equivalent of B<--diagrams=none> is
+B<--define=diagrams:show=0>.
 
 B<--define> may be used multiple times to set multiple items.
 
@@ -434,14 +482,14 @@ configuration.
 
 =item B<--print-default-config>
 
-Prints the default configuration, and exits.
+Prints the default configuration to standard output, and exits.
 
 The default configuration is commented to explain its contents.
 
 =item B<--print-final-config>
 
 Prints the final configuration (after processing all system, user and
-other config files), and exits.
+other config files)  to standard output, and exits.
 
 The final configuration is not commented. Sorry.
 
@@ -453,11 +501,11 @@ The final configuration is not commented. Sorry.
 
 =item B<--help> (short: -h)
 
-Prints help message. No other output is produced.
+Prints a help message. No other output is produced.
 
 =item B<--manual>
 
-Prints the manual. No other output is produced.
+Prints the manual page. No other output is produced.
 
 =item B<--ident>
 
@@ -572,23 +620,27 @@ sub app_setup {
           "lyrics-only",                # Suppress all chords
           "generate=s",
           "backend-option|bo=s\%",
+	  "diagrams=s",			# Prints chord diagrams
           "encoding=s",
-          "user-chord-grids!",          # Do[esn't] print grids for user defined chords.
+	  "csv!",			# Generates contents CSV
+	  "cover=s",			# Cover page(s)
 
           ### Standard Chordii Options ###
 
           "about|A" => \$about,         # About...
           "chord-font|C=s",             # Sets chord font
-          "chord-grid-size|s=f",        # Sets chord grid size [30]
-          "chord-grids-sorted|S!",      # Prints chord grids ordered
+          "chord-grid-size|s=f",        # Sets chord diagram size [30]
+          "chord-grids-sorted|S!",      # Prints chord diagrams ordered
           "chord-size|c=i",             # Sets chord size [9]
+	  "diagrams=s",			# Prints chords diagrams
           "dump-chords|D",              # Dumps chords definitions (PostScript)
           "dump-chords-text|d" => \$dump_chords,  # Dumps chords definitions (Text)
           "even-pages-number-left|L",   # Even pages numbers on left
           "odd-pages-number-left",      # Odd pages numbers on left
           "lyrics-only|l",              # Only prints lyrics
-          "chord-grids|G!",             # En[dis]ables printing of chord grids
-          "easy-chord-grids|g!",        # Do[esn't] print grids for built-in "easy" chords.
+          "G" => sub { $clo->{'chord-grids'} = 0 },
+          "chord-grids!",               # En[dis]ables printing of chord diagrams
+          "easy-chord-grids|g!",        # Do[esn't] print diagrams for built-in "easy" chords. Ignored.
           "page-number-logical|n",      # Numbers logical pages, not physical
           "page-size|P=s",              # Specifies page size [letter, a4 (default)]
           "single-space|a!",            # Automatic single space lines without chords
@@ -598,6 +650,7 @@ sub app_setup {
           "i" => sub { $clo->{toc} = 1 },
           "toc!",                       # Generates a table of contents
           "transpose|x=i",              # Transposes by N semi-tones
+          "user-chord-grids!",          # Do[esn't] print diagrams for user defined chords.
           "version|V" => \$version,     # Prints version and exits
           "vertical-space|w=f",         # Extra vertical space between lines
           "2-up|2",                     # 2 pages per sheet
@@ -637,9 +690,14 @@ sub app_setup {
         # Load Pod::Usage only if needed.
         require Pod::Usage;
         Pod::Usage->import;
-        unshift( @_,
-		 -input => App::Packager::GetResource
-		 ( $manual == 2 ? "pod/Config.pod" : "pod/ChordPro.pod" ) );
+	my $f = $manual == 2 ? "pod/Config.pod" : "pod/ChordPro.pod";
+	if ( $App::Packager::PACKAGED ) {
+	    $f = App::Packager::GetResource($f);
+	}
+	else {
+	    $f = ::findlib($f);
+	}
+        unshift( @_, -input => $f );
         &pod2usage;
     };
 
@@ -670,8 +728,13 @@ sub app_setup {
     for my $config ( qw(config) ) {
         for ( $clo->{$config} ) {
             if ( defined($_) ) {
-                foreach ( @$_ ) {
-                    die("$_: $!\n") unless -r $_;
+                foreach my $c ( @$_ ) {
+		    # Check for resource names.
+		    if ( ! -r $c && $c !~ m;[/.]; ) {
+			my $t = ::findlib( "config/".lc($c).".json" );
+			$c = $t if $t;
+		    }
+                    die("$c: $!\n") unless -r $c;
                 }
                 next;
             }
@@ -705,6 +768,10 @@ sub app_setup {
 	exit 0;
     }
 
+    # At this point, there should be filename argument(s)
+    # unless we're embedded.
+    app_usage(\*STDERR, 1) unless $::__EMBEDDED__ || @ARGV;
+
     # Return result.
     $options;
 }
@@ -749,11 +816,17 @@ EndOfAbout
 
 sub app_usage {
     my ($fh, $exit) = @_;
+    my %cfg;
+    for ( qw( config userconfig sysconfig) ) {
+	$cfg{$_} = $configs{$_} || "no default";
+    }
     print ${fh} <<EndOfUsage;
 Usage: $0 [ options ] [ file ... ]
 
 Options:
     --about  -A                   About ChordPro...
+    --cover=FILE                  Add cover pages from PDF document
+    --diagrams=WHICH		  Prints chord diagrams
     --encoding=ENC                Encoding for input files (UTF-8)
     --lyrics-only  -l             Only prints lyrics
     --output=FILE  -o             Saves the output to FILE
@@ -761,37 +834,37 @@ Options:
     --start-page-number=N  -p     Starting page number [1]
     --toc --notoc -i              Generates/suppresses a table of contents
     --transpose=N  -x             Transposes by N semi-tones
-    --user-chord-grids		  Prints the user defined chords in the song
     --version  -V                 Prints version and exits
 
 Chordii compatibility.
 Options marked with * are better specified in the config file.
 Options marked with - are ignored.
     --chord-font=FONT  -C         *Sets chord font
-    --chord-grid-size=N  -s       *Sets chord grid size [30]
-    --chord-grids-sorted  -S      *Prints chord grids ordered by key
+    --chord-grid-size=N  -s       *Sets chord diagram size [30]
+    --chord-grids-sorted  -S      *Prints chord diagrams ordered by key
     --chord-size=N  -c            *Sets chord size [9]
     --dump-chords  -D             Dumps chords definitions (PostScript)
     --dump-chords-text  -d        Dumps chords definitions (Text)
     --even-pages-number-left  -L  *Even pages numbers on left
     --odd-pages-number-left       *Odd pages numbers on left
-    --no-chord-grids  -G          *Disables printing of chord grids
+    --no-chord-grids  -G          *Disables printing of chord diagrams
     --no-easy-chord-grids  -g     Not supported
     --page-number-logical  -n     -Numbers logical pages, not physical
     --page-size=FMT  -P           *Specifies page size [letter, a4 (default)]
     --single-space  -a            *Automatic single space lines without chords
     --text-size=N  -t             *Sets text size [12]
     --text-font=FONT  -T          *Sets text font
+    --user-chord-grids		  *Prints the user defined chords in the song
     --vertical-space=N  -w        *Extra vertical space between lines
     --2-up  -2                    Not supported
     --4-up  -4                    Not supported
 
 Configuration options:
-    --config=CFG        Project specific config file ($configs{config})
+    --config=CFG        Project specific config file ($cfg{config})
     --noconfig          Don't use a project specific config file
-    --userconfig=CFG    User specific config file ($configs{userconfig})
+    --userconfig=CFG    User specific config file ($cfg{userconfig})
     --nouserconfig      Don't use a user specific config file
-    --sysconfig=CFG     System specific config file ($configs{sysconfig})
+    --sysconfig=CFG     System specific config file ($cfg{sysconfig})
     --nosysconfig       Don't use a system specific config file
     --nolegacyconfig    Don't use a Chord/Chordii legacy config file
     --nodefaultconfigs  -X  Don't use any default config files
@@ -810,15 +883,41 @@ EndOfUsage
     exit $exit if defined $exit;
 }
 
+################ Resources ################
+
+sub ::findlib {
+    my ( $file ) = @_;
+
+    # Packaged.
+    if ( $App::Packager::PACKAGED ) {
+	my $found = App::Packager::GetUserFile($file);
+	return $found if -e $found;
+	$found = App::Packager::GetResource($file);
+	return $found if -e $found;
+    }
+
+    ( my $me = __PACKAGE__ ) =~ s;::;/;g;
+    foreach ( @INC ) {
+	return "$_/$me/user/$file" if -e "$_/$me/user/$file";
+	return "$_/$me/res/$file"  if -e "$_/$me/res/$file";
+	return "$_/$me/$file"      if -e "$_/$me/$file";
+    }
+    undef;
+}
+
+use lib ( grep { defined } ::findlib("CPAN") );
+
 =head1 FONTS
 
 There are two ways to specify fonts: with a font filename, and a
 built-in font name.
 
 A font filename must be either and absolute filename, or a relative
-filename which is interpreted relative to the configuration setting
-C<fontdir>. In any case, the filename should point to a valid TrueType
-(C<.ttf>) or OpenType (C<.otf>) font.
+filename which is interpreted relative to the I<font path>, which
+consists of configuration setting C<fontdir>, the C<fonts> resource
+dir, and the contents of environment variable C<FONTDIR>. In any case,
+the filename should point to a valid TrueType (C<.ttf>) or OpenType
+(C<.otf>) font.
 
 If it is not a filename, it must be the name one of the built-in fonts.
 
@@ -833,7 +932,7 @@ Built-in 'Adobe Core Fonts':
   Helvetica-BoldOblique
   Helvetica-Oblique
 
-Buitl-in 'Windows Fonts':
+Built-in 'Windows Fonts':
 
   Georgia                             Webdings
   Georgia,Bold                        Wingdings
@@ -871,6 +970,8 @@ supports most of the features of Chordii, and a lot more:
 
 * User defined chords and tuning, not limited to 6 strings.
 
+* Support for Nashville Numbering and Roman Numbering.
+
 * Support for external TrueType and OpenType fonts
 
 * Font kerning (with external TrueType fonts)
@@ -881,11 +982,23 @@ supports most of the features of Chordii, and a lot more:
 
 (* = under development)
 
+=head1 AUTHOR
+
+Johan Vromans C<< <jv at CPAN dot org > >>
+
+=head1 SUPPORT
+
+ChordPro (the program) development is hosted on GitHub, repository
+L<https://github.com/sciurius/chordpro>.
+
+Please report any bugs or feature requests to the GitHub issue tracker,
+L<https://github.com/sciurius/chordpro/issues>.
+
 =head1 LICENSE
 
-Copyright (C) 2010,2016 Johan Vromans,
+Copyright (C) 2010,2017 Johan Vromans,
 
-This module is free software. You can redistribute it and/or
+This program is free software. You can redistribute it and/or
 modify it under the terms of the Artistic License 2.0.
 
 This program is distributed in the hope that it will be useful,
