@@ -33,10 +33,10 @@ This is the current built-in configuration file, showing all settings.
   	"columns" : 1,
   	// Suppress empty chord lines.
   	// Overrides the -a (--single-space) command line options.
-  	"suppress-empty-chords" : 1,
+  	"suppress-empty-chords" : true,
   	// Suppress chords.
   	// Overrides --lyrics-only command line option.
-  	"lyrics-only" : 0,
+  	"lyrics-only" : false,
       },
   
       // Metadata.
@@ -47,9 +47,10 @@ This is the current built-in configuration file, showing all settings.
       // The separator is used to concatenate multiple values.
       "metadata" : {
         "keys" : [ "title", "subtitle",
-                   "artist", "composer", "album",
-                   "key", "time", "tempo", "capo" ],
-        "strict" : 1,
+                   "artist", "composer", "lyricist", "arranger",
+                   "album", "copyright", "year",
+                   "key", "time", "tempo", "capo", "duration" ],
+        "strict" : true,
         "separator" : "; ",
       },
   
@@ -67,20 +68,21 @@ This is the current built-in configuration file, showing all settings.
   	//    "name"  : "Bb",
   	//    "base"  : 1,
   	//    "frets" : [ 1, 1, 3, 3, 3, 1 ],
-  	//    "easy"  : 1,
+  	//    "fingers" : [ 1, 1, 2, 3, 4, 1 ],
+  	//    "easy"  : true,
   	//  },
       ],
   
-      // Printing chord grids.
-      // "auto": automatically add unknown chords as empty grids.
+      // Printing chord diagrams.
+      // "auto": automatically add unknown chords as empty diagrams.
       // "show": prints the chords used in the song.
       //         "all": all chords used.
       //         "user": only prints user defined chords.
       // "sorted": order the chords by key.
-      "chordgrid" : {
-          "auto"     :  0,
+      "diagrams" : {
+          "auto"     :  false,
           "show"     :  "all",
-          "sorted"   :  0,
+          "sorted"   :  false,
       },
   
       // Diagnostig messages.
@@ -109,7 +111,7 @@ This is the current built-in configuration file, showing all settings.
   
   	// Special: head on first page only, add the headspace to
 	// the other pages so they become larger.
-  	"head-first-only" : 0,
+  	"head-first-only" : false,
   
   	// Spacings.
   	// Baseline distances as a factor of the font size.
@@ -141,7 +143,7 @@ This is the current built-in configuration file, showing all settings.
             "recall" : {
                  "tag"   : "Chorus",
                  "type"  : "comment",
-                 "quote" : 0,
+                 "quote" : false,
             },
 	},
   
@@ -152,22 +154,24 @@ This is the current built-in configuration file, showing all settings.
   
 	// A {titles: left} may conflict with customized formats.
 	// Set to non-zero to ignore the directive.
-	"titles-directive-ignore" : 0,
+	"titles-directive-ignore" : false,
   
-    	// Chord grids.
-  	// A chord grid consists of a number of cells.
+    	// Chord diagrams.
+  	// A chord diagram consists of a number of cells.
   	// Cell dimensions are specified by "width" and "height".
   	// The horizontal number of cells depends on the number of strings.
   	// The vertical number of cells is "vcells", which should
 	// be 4 or larger to accomodate most chords.
-  	// The horizontal distance between grids is "hspace" cells.
+  	// The horizontal distance between diagrams is "hspace" cells.
   	// The vertical distance is "vspace" cells.
-  	"chordgrid" : {
+	// "linewidth" is the thickness of the lines as a fraction of "width".
+  	"diagrams" : {
 	    "width"    :  6,
   	    "height"   :  6,
   	    "hspace"   :  3.95,
   	    "vspace"   :  3,
   	    "vcells"   :  4,
+	    "linewidth" : 0.1,
   	},
   
   	// Even/odd pages. A value of -1 denotes odd/even pages.
@@ -243,6 +247,10 @@ This is the current built-in configuration file, showing all settings.
   		"name" : "Times-Roman",
   		"size" : 11
   	    },
+  	    "grid" : {
+  		"name" : "Helvetica",
+  		"size" : 10
+  	    },
   	},
   
   	// Fonts that can be specified, but need not.
@@ -251,14 +259,15 @@ This is the current built-in configuration file, showing all settings.
   	// comment_italic --> chord
   	// comment_box    --> chord
   	// toc            --> text
-  	// grid           --> comment
+  	// grid           --> chord
+  	// grid_margin	  --> comment
   	// footer         --> subtitle @ 60%
   	// empty          --> text
-	// chordgrid	  --> comment
-	// chordgrid_capo --> text (but at a small size)
+	// diagram	  --> comment
+	// diagram_base	  --> text (but at a small size)
   
   	// This will show the page layout if non-zero.
-  	"showlayout" : 0,
+  	"showlayout" : false,
       },
   }
   // End of config.
@@ -323,11 +332,15 @@ sub configurator {
     for ( qw(tuning) ) {
 	$cfg->{$_} //= undef;
     }
+    for ( qw(title subtitle footer) ) {
+	next if exists($cfg->{pdf}->{formats}->{first}->{$_});
+	$cfg->{pdf}->{formats}->{first}->{$_} = "";
+    }
     for my $ff ( qw(chord
-		    chordgrid chordgrid_capo
+		    diagram diagram_capo
 		    comment comment_box comment_italic
 		    tab text toc
-		    empty footer grid subtitle title) ) {
+		    empty footer grid grid_margin subtitle title) ) {
 	for ( qw(name file size color background) ) {
 	    $cfg->{pdf}->{fonts}->{$ff}->{$_} //= undef;
 	}
@@ -377,6 +390,19 @@ sub configurator {
 	delete( $cfg->{$_} )
 	  unless defined( $cfg->{$_} );
     }
+    for ( qw(title subtitle footer) ) {
+	delete($cfg->{pdf}->{formats}->{first}->{$_})
+	  if ($cfg->{pdf}->{formats}->{first}->{$_} // 1) eq "";
+	for my $class ( qw(title first default) ) {
+	    my $t = $cfg->{pdf}->{formats}->{$class}->{$_};
+	    next unless $t;
+	    die("Config error in pdf.formats.$class.$_: not an array\n")
+	      unless ref($t) eq 'ARRAY';
+	    die("Config error in pdf.formats.$class.$_: ",
+		 scalar(@$t), " fields instead of 3\n")
+	      unless @$t == 3;
+	}
+    }
     my @allfonts = keys(%{$cfg->{pdf}->{fonts}});
     for my $ff ( @allfonts ) {
 	unless ( $cfg->{pdf}->{fonts}->{$ff}->{name}
@@ -390,16 +416,22 @@ sub configurator {
 	}
     }
 
-    if ( defined $options->{'user-chord-grids'} ) {
-	$cfg->{chordgrid}->{show} =
+    if ( defined $options->{diagrams} ) {
+	warn( "Invalid value for diagrams: ",
+	      $options->{diagrams}, "\n" )
+	  unless $options->{diagrams} =~ /^(all|none|user)$/i;
+	$cfg->{diagrams}->{show} = lc $options->{'diagrams'};
+    }
+    elsif ( defined $options->{'user-chord-grids'} ) {
+	$cfg->{diagrams}->{show} =
 	  $options->{'user-chord-grids'} ? "user" : 0;
     }
     elsif ( defined $options->{'chord-grids'} ) {
-	$cfg->{chordgrid}->{show} =
+	$cfg->{diagrams}->{show} =
 	  $options->{'chord-grids'} ? "all" : 0;
     }
     if ( defined $options->{'chord-grids-sorted'} ) {
-	$cfg->{chordgrid}->{sorted} = $options->{'chord-grids-sorted'};
+	$cfg->{diagrams}->{sorted} = $options->{'chord-grids-sorted'};
     }
 
     if ( $cfg->{tuning} ) {
@@ -412,7 +444,7 @@ sub configurator {
     foreach ( @{ $cfg->{chords} } ) {
 	my $res =
 	  App::Music::ChordPro::Chords::add_config_chord
-	      ( $_->{name}, $_->{base}||1, $_->{frets}, $_->{easy} );
+	      ( $_->{name}, $_->{base}||1, $_->{frets}, $_->{easy}, $_->{fingers} );
 	warn( "Invalid chord in config: ",
 	      $_->{name}, ": ", $res, "\n" ) if $res;
     }
@@ -472,14 +504,15 @@ sub add_legacy {
 	    $cfg->{settings}->{columns} = $song->{settings}->{columns};
 	    next;
 	}
-	if ( $_ eq "showgrids" ) {
-	    $cfg->{chordgrid}->{show} = $song->{settings}->{showgrids};
+	if ( $_ eq "diagrams" ) {
+	    $cfg->{diagrams}->{show} = $song->{settings}->{diagrams};
 	    next;
 	}
 	die("Cannot happen");
     }
     foreach ( @{$song->{body}} ) {
-	next if $_->{type} eq "chord-grids"; # added by parser
+	next if $_->{type} eq "diagrams"; # added by parser
+	next if $_->{type} eq "ignore"; # ignored
 	unless ( $_->{type} eq "control" ) {
 	    die("Cannot happen " . $_->{type} . " " . $_->{name});
 	}
@@ -542,24 +575,6 @@ sub hmerge($$$) {
     return \%res;
 }
 
-=begin later
-
-my $cloner;
-
-sub clone($) {
-    my $self = shift;
-
-    unless ( $cloner ) {
-	eval { require Clone; $cloner = \&Clone::clone }
-	  or
-	eval { require Clone::PP; $cloner = \&Clone::PP::clone }
-    }
-    my $h = $cloner->($self);
-    bless( $h, ref($self) );
-}
-
-=cut
-
 sub clone($) {
     my ( $source ) = @_;
 
