@@ -11,6 +11,15 @@ use Encode qw( encode_utf8 );
 
 use App::Music::ChordPro::Output::Common;
 
+my $pdfapi;
+BEGIN {
+    eval { require PDF::Builder; $pdfapi = "PDF::Builder"; }
+      or
+    eval { require PDF::API2; $pdfapi = "PDF::API2"; }
+      or
+    die("Missing PDF::API package\n");
+}
+
 use constant DEBUG_SPACING => 0;
 
 # For regression testing, run perl with PERL_HASH_SEED set to zero.
@@ -24,7 +33,7 @@ sub generate_songbook {
     return [] unless $sb->{songs}->[0]->{body}; # no songs
 
     my $ps = $::config->{pdf};
-    my $pr = PDFWriter->new( $ps, $options->{output} || "__new__.pdf" );
+    my $pr = PDFWriter->new($ps);
     $pr->info( Title => $sb->{songs}->[0]->{meta}->{title}->[0],
 	       Creator =>
 	       $regtest
@@ -42,44 +51,65 @@ sub generate_songbook {
 				{ pr => $pr, $options ? %$options : () } );
     }
 
-    if ( $options->{toc} ) {
-
-	$pr->newpage($ps), $page++
-	  if $ps->{'even-odd-pages'} && $page % 2 == 0;
+    if ( $options->{toc} // @book > 1 ) {
 
 	# Create a pseudo-song for the table of contents.
-	$options->{startpage} = $page;
+	my $t = get_format( $ps, 1, "toc-title" );
 	my $song =
-	  { title     => get_format( $ps, 1, "toc-title" ),
+	  { title     => $t,
+	    meta => { title => [ $t ] },
 	    structure => "linear",
 	    body      => [
 		     map { +{ type    => "tocline",
 			      context => "toc",
 			      title   => $_->[0],
-			      page    => $_->[1],
+			      pageno  => $_->[1],
+			      page    => $pr->{pdf}->openpage($_->[1]),
 			    } } @book,
 	    ],
-	    meta      => {},
 	  };
-	push( @book, [ $song->{title}, $page ] );
-	$page += generate_song( $song,
-				{ pr => $pr, $options ? %$options : () } );         }
 
-    $pr->finish;
+	# Prepend the toc.
+	$options->{startpage} = 1;
+	$page = generate_song( $song,
+			       { pr => $pr, prepend => 1,
+				 $options ? %$options : () } );
 
-    if ( $options->{toc} ) {
+	# Align.
+	$pr->newpage($ps, $page+1), $page++
+	  if $ps->{'even-odd-pages'} && $page % 2;
+    }
+    else {
+	$page = 1;
+    }
+
+    if ( $options->{cover} ) {
+	my $cover = $pdfapi->open( $options->{cover} );
+	die("Missing cover: ", $options->{cover}, "\n") unless $cover;
+	for ( 1 .. $cover->pages ) {
+	    $pr->{pdf}->importpage( $cover, $_, $_ );
+	    $page++;
+	}
+	$pr->newpage( $ps, 1+$cover->pages ), $page++
+	  if $ps->{'even-odd-pages'} && $page % 2;
+    }
+
+    $pr->finish( $options->{output} || "__new__.pdf" );
+
+    if ( $options->{csv} ) {
 
 	# Create an MSPro compatible CSV for this PDF.
+	push( @book, [ "CSV", $page ] );
 	( my $csv = $options->{output} ) =~ s/\.pdf$/.csv/i;
 	open( my $fd, '>:utf8', encode_utf8($csv) )
 	  or die( encode_utf8($csv), ": $!\n" );
 	print $fd ( "title;pages;\n" );
-	for ( my $p = 1; $p < @book-1; $p++ ) {
+	for ( my $p = 0; $p < @book-1; $p++ ) {
 	    print $fd ( join(';',
 			     $book[$p]->[0],
 			     $book[$p+1]->[1] > $book[$p]->[1]+1
-			     ? ( $book[$p]->[1] ."-". ($book[$p+1]->[1]-1) )
-			     : $book[$p]->[1]),
+			     ? ( $page+$book[$p]->[1] ."-". ($page+$book[$p+1]->[1]-1) )
+			     : $page+$book[$p]->[1]),
 			"\n" );
 	}
 	close($fd);
@@ -94,7 +124,7 @@ my $lyrics_only = 0;		# suppress all chord lines
 my $chordscol = 0;		# chords in a separate column
 my $chordscapo = 0;		# capo in a separate column
 
-use constant SIZE_ITEMS => [ qw (chord text tab grid toc title footer) ];
+use constant SIZE_ITEMS => [ qw (chord text tab grid diagram toc title footer) ];
 
 sub generate_song {
     my ($s, $options) = @_;
@@ -147,7 +177,7 @@ sub generate_song {
 	$fonts->{$_}->{_size} = $fonts->{$_}->{size};
     }
 
-    my $x = $ps->{marginleft} + $ps->{columnoffsets}->[0];
+    my $x;
     my $y = $ps->{papersize}->[1] - $ps->{margintop};
 
     $ps->{'even-odd-pages'} =  1 if $options->{'even-pages-number-left'};
@@ -160,7 +190,11 @@ sub generate_song {
 	    my ( $from, $to ) = @_;
 	    for my $class ( qw( default title first ) ) {
 		for ( qw( title subtitle footer ) ) {
-		    next unless exists $ps->{formats}->{$class}->{$_};
+		    next unless defined $ps->{formats}->{$class}->{$_};
+		    unless ( ref($ps->{formats}->{$class}->{$_}) eq 'ARRAY' ) {
+			warn("Oops -- pdf.formats.$class.$_ is not an array\n");
+			next;
+		    }
 		    ( $ps->{formats}->{$class}->{$_}->[$from],
 		      $ps->{formats}->{$class}->{$_}->[$to] ) =
 			( $ps->{formats}->{$class}->{$_}->[$to],
@@ -191,6 +225,22 @@ sub generate_song {
     };
 
     my $col;
+
+    my $col_adjust = sub {
+	return if $ps->{columns} <= 1;
+	$x = $ps->{_leftmargin} + $ps->{columnoffsets}->[$col];
+	$ps->{__leftmargin} = $x;
+	$ps->{__rightmargin} =
+	  $ps->{_leftmargin}
+	    + $ps->{columnoffsets}->[$col+1];
+	$ps->{__rightmargin} -= $ps->{columnspace}
+	  if $col < $ps->{columns}-1;
+	warn("C=$col, L=", $ps->{__leftmargin},
+	     ", R=", $ps->{__rightmargin},
+	     "\n") if DEBUG_SPACING;
+	$y = $ps->{papersize}->[1] - $ps->{margintop};
+    };
+
     my $vsp_ignorefirst;
     my $startpage = $options->{startpage} || 1;
     my $thispage = $startpage - 1;
@@ -199,24 +249,31 @@ sub generate_song {
     my $newpage = sub {
 
 	# Add page to the PDF.
-	$pr->newpage($ps);
-
-	# Prepare for printing.
-	showlayout($ps) if $ps->{showlayout};
+	$pr->newpage($ps, $options->{prepend} ? $thispage+1 : () );
 
 	# Put titles and footer.
 
 	# If even/odd pages, leftpage signals whether the
 	# header/footer parts must be swapped.
-	my $leftpage;
+	my $rightpage = 1;
 	if ( $ps->{"even-odd-pages"} ) {
 	    # Even/odd printing...
-	    $leftpage = $thispage % 2 != 0;
+	    $rightpage = $thispage % 2 == 0;
 	    # Odd/even printing...
-	    $leftpage = !$leftpage if $ps->{'even-odd-pages'} < 0;
+	    $rightpage = !$rightpage if $ps->{'even-odd-pages'} < 0;
 	}
+	if ( $rightpage ) {
+	    $ps->{_leftmargin}  = $ps->{marginleft};
+	    $ps->{_rightmargin} = $ps->{marginright};
+	}
+	else {
+	    $ps->{_leftmargin}  = $ps->{marginright};
+	    $ps->{_rightmargin} = $ps->{marginleft};
+	}
+	showlayout($ps) if $ps->{showlayout};
 
 	$thispage++;
+	$s->{meta}->{page} = [ $s->{page} = $thispage ];
 
 	# Determine page class.
 	my $class = 2;		# default
@@ -226,12 +283,13 @@ sub generate_song {
 	elsif ( $thispage == $startpage ) {
 	    $class = 1;		# first of a song
 	}
-	$s->{page} = $thispage;
 
 	# Three-part title handlers.
-	my $tpt = sub { tpt( $ps, $class, $_[0], $leftpage, $x, $y, $s ) };
+	my $tpt = sub { tpt( $ps, $class, $_[0], $rightpage, $x, $y, $s ) };
+
+	$ps->{__leftmargin} = $x = $ps->{_leftmargin};
+	$ps->{__rightmargin} = $ps->{papersize}->[0] - $ps->{_rightmargin};
 
-	$x = $ps->{marginleft};
 	if ( $ps->{headspace} ) {
 	    $y = $ps->{papersize}->[1] - $ps->{margintop} + $ps->{headspace};
 	    $y -= font_bl($fonts->{title});
@@ -253,6 +311,7 @@ sub generate_song {
 	$y += $ps->{headspace} if $ps->{'head-first-only'} && $class == 2;
 	$col = 0;
 	$vsp_ignorefirst = 1;
+	$col_adjust->();
     };
 
     # Get going.
@@ -272,11 +331,7 @@ sub generate_song {
 	    $newpage->();
 	    $vsp_ignorefirst = 0;
 	}
-	else {
-	    $x = $ps->{marginleft} + $ps->{columnoffsets}->[$col];
-	    $y = $ps->{papersize}->[1] - $ps->{margintop};
-	}
-
+	$col_adjust->();
 	return;
     };
 
@@ -284,11 +339,10 @@ sub generate_song {
     my $elt;			# current element
 
     my $prev;			# previous element
-    my @chorus;			# chorus elements, if any
 
     my $grid_cellwidth;
-    my $grid_barwidth = 8;	# tentative
-    $grid_barwidth *= 1.5;		#####
+    my $grid_barwidth = 0.5 * $fonts->{chord}->{size};
+    my $grid_margin;
 
     while ( @elts ) {
 	$elt = shift(@elts);
@@ -315,12 +369,6 @@ sub generate_song {
 	    next;
 	}
 
-	# Collect chorus elements so they can be recalled.
-	if ( $elt->{context} eq "chorus" ) {
-	    @chorus = () unless $prev && $prev->{context} eq "chorus";
-	    push( @chorus, $elt );
-	}
-
 	if ( $elt->{type} eq "songline"
 	     or $elt->{type} eq "tabline"
 	     or $elt->{type} =~ /^comment(?:_box|_italic)?$/ ) {
@@ -355,10 +403,9 @@ sub generate_song {
 		my $style = $ps->{chorus};
 		$indent = $style->{indent};
 		if ( $style->{bar}->{offset} && $style->{bar}->{width} ) {
-		    my $cx = $ps->{marginleft}
-		      + $ps->{columnoffsets}->[$col]
-			- $style->{bar}->{offset}
-			  + $indent;
+		    my $cx = $ps->{__leftmargin}
+		      - $style->{bar}->{offset}
+			+ $indent;
 		    $pr->vline( $cx, $y, $vsp,
 				$style->{bar}->{width},
 				$style->{bar}->{color} );
@@ -368,7 +415,14 @@ sub generate_song {
 	    # Substitute metadata in comments.
 	    if ( $elt->{type} =~ /^comment/ ) {
 		$elt = { %$elt };
-		$elt->{text} = fmt_subst( $s, $elt->{text}, 1 );
+		# Flatten chords/phrases.
+		if ( $elt->{chords} ) {
+		    $elt->{text} = "";
+		    for ( 0..$#{ $elt->{chords} } ) {
+			$elt->{text} .= $elt->{chords}->[$_] . $elt->{phrases}->[$_];
+		    }
+		}
+		$elt->{text} = fmt_subst( $s, $elt->{text} );
 	    }
 
 	    # Comment decorations.
@@ -418,8 +472,7 @@ sub generate_song {
 		}
 	    }
 	    my $style = $ps->{chorus};
-	    my $cx = $ps->{marginleft}
-	      + $ps->{columnoffsets}->[$col] - $style->{bar}->{offset};
+	    my $cx = $ps->{__leftmargin} - $style->{bar}->{offset};
 	    $pr->vline( $cx, $cy, vsp($ps), 1, $style->{bar}->{color} );
 	    $y -= vsp($ps,4); # chordii
 	    next;
@@ -449,8 +502,21 @@ sub generate_song {
 	    $checkspace->($vsp);
 	    $pr->show_vpos( $y, 0 ) if DEBUG_SPACING;
 
+	    my $cells = $grid_margin->[2];
+	    $grid_cellwidth = ( $ps->{__rightmargin}
+				- $ps->{__leftmargin}
+				- ($cells)*$grid_barwidth
+			      ) / $cells;
+	    warn("L=", $ps->{__leftmargin},
+		 ", R=", $ps->{__rightmargin},
+		 ", C=$cells, W=", $grid_cellwidth,
+		 "\n") if DEBUG_SPACING;
+
 	    gridline( $elt, $x, $y,
-		      $grid_cellwidth, $grid_barwidth, $ps );
+		      $grid_cellwidth,
+		      $grid_barwidth,
+		      $grid_margin,
+		      $ps );
 
 	    $y -= $vsp;
 	    $pr->show_vpos( $y, 1 ) if DEBUG_SPACING;
@@ -517,7 +583,7 @@ sub generate_song {
 	if ( $elt->{type} eq "rechorus" ) {
 	    my $t = $ps->{chorus}->{recall};
 	    if ( $t->{quote} ) {
-		unshift( @elts, @chorus );
+		unshift( @elts, @{ $elt->{chorus} } ) if $elt->{chorus};
 	    }
 	    if ( $t->{tag} && $t->{type} =~ /^comment(?:_(?:box|italic))?/ ) {
 		unshift( @elts, { %$elt,
@@ -540,20 +606,19 @@ sub generate_song {
 	    next;
 	}
 
-	if ( $elt->{type} eq "chord-grids" ) {
+	if ( $elt->{type} eq "diagrams" ) {
 
 	    my @chords = @{ $elt->{chords} };
 
-	    local $::config->{chordgrid}->{show} =
-	      $elt->{show} || $::config->{chordgrid}->{show};
+	    local $::config->{diagrams}->{show} =
+	      $elt->{show} || $::config->{diagrams}->{show};
 
 	    my $vsp = chordgrid_vsp( $elt, $ps );
 	    my $hsp = chordgrid_hsp( $elt, $ps );
-	    my $h = int( ( $ps->{papersize}->[0]
-			   - $ps->{marginleft}
-			   - $ps->{marginright}
-			   + $ps->{chordgrid}->{hspace}
-			     * $ps->{chordgrid}->{width} ) / $hsp );
+	    my $h = int( ( $ps->{__rightmargin}
+			   - $ps->{__leftmargin}
+			   + $ps->{diagrams}->{hspace}
+			   * $ps->{diagrams}->{width} ) / $hsp );
 	    while ( @chords ) {
 		my $x = $x;
 		$checkspace->($vsp);
@@ -574,20 +639,37 @@ sub generate_song {
 
 	if ( $elt->{type} eq "control" ) {
 	    if ( $elt->{name} =~ /^(text|chord|grid|toc|tab)-size$/ ) {
-		$do_size->( $1, $elt->{value} );
+		if ( defined $elt->{value} ) {
+		    $do_size->( $1, $elt->{value} );
+		}
+		else {
+		    warn("PDF: No value for property $1-size\n" );
+		}
 	    }
 	    elsif ( $elt->{name} =~ /^(text|chord|grid|toc|tab)-font$/ ) {
 		my $f = $1;
-		if ( $elt->{value} =~ m;/; ) {
-		    $ps->{fonts}->{$1}->{file} = $elt->{value};
+		if ( defined $elt->{value} ) {
+		    if ( $elt->{value} =~ m;/; ) {
+			delete $ps->{fonts}->{$f}->{name};
+			$ps->{fonts}->{$f}->{file} = $elt->{value};
+		    }
+		    else {
+			delete $ps->{fonts}->{$f}->{file};
+			$ps->{fonts}->{$f}->{name} = $elt->{value};
+		    }
+		    $pr->init_font($f);
 		}
 		else {
-		    $ps->{fonts}->{$1}->{name} = $elt->{value};
+		    warn("PDF: No value for property $1-font\n" );
 		}
-		$pr->init_font($ps->{fonts}->{$f});
 	    }
 	    elsif ( $elt->{name} =~ /^(text|chord|grid|toc|tab)-color$/ ) {
-		$ps->{fonts}->{$1}->{color} = $elt->{value};
+		if ( defined $elt->{value} ) {
+		    $ps->{fonts}->{$1}->{color} = $elt->{value};
+		}
+		else {
+		    warn("PDF: No value for property $1-color\n" );
+		}
 	    }
 	    next;
 	}
@@ -601,6 +683,7 @@ sub generate_song {
 		my @v = @{ $elt->{value} };
 		my $cells;
 		my $bars = 8;
+		$grid_margin = [ 0, 0 ];
 		if ( $v[1] ) {
 		    $cells = $v[0] * $v[1];
 		    $bars = $v[0];
@@ -608,12 +691,24 @@ sub generate_song {
 		else {
 		    $cells = $v[0];
 		}
-		$grid_cellwidth = ( $ps->{papersize}->[0]
-				    - $ps->{marginleft}
-				    - $ps->{marginright}
-				    - (1+$bars)*$grid_barwidth
-				  ) / $cells;
+		$cells += $grid_margin->[0] = $v[2] if $v[2];
+		$cells += $grid_margin->[1] = $v[3] if $v[3];
+		$grid_margin->[2] = $cells;
 	    }
+	    # Arbitrary config values.
+	    elsif ( $elt->{name} =~ /^pdf\.(.+)/ ) {
+		my @k = split( /[.]/, $1 );
+		my $cc = {};
+		my $c = \$cc;
+		foreach ( @k ) {
+		    $c = \($$c->{$_});
+		}
+		$$c = $elt->{value};
+		$ps = App::Music::ChordPro::Config::hmerge( $ps, $cc, "" );
+	    }
+	    next;
+	}
+	if ( $elt->{type} eq "ignore" ) {
 	    next;
 	}
 
@@ -779,7 +874,36 @@ sub songline {
 	    push(@chords, $chord);
 	}
 	else {
-	    my $xt0 = $pr->text( $chord." ", $x, $ychord, $fchord );
+	    my $info = App::Music::ChordPro::Chords::identify($chord);
+	    my $xt0;
+	    if ( $info && $info->{system} eq "R" ) {
+		$xt0 = $pr->text( $info->{dproot}.$info->{qual},
+				  $x, $ychord, $fchord );
+		$xt0 = $pr->text( $info->{adds}, $xt0,
+				   $ychord + $fchord->{size} * 0.2,
+				   $fchord,
+				   $fchord->{size} * 0.8
+				 );
+		$xt0 = $pr->text( " ", $xt0, $ychord, $fchord );
+	    }
+	    elsif ( $info && $info->{system} eq "N" ) {
+		$xt0 = $pr->text( $info->{dproot}.$info->{qual},
+				  $x, $ychord, $fchord );
+#		if ( $info->{minor} ) {
+#		    my $m = $info->{minor};
+#		    # $m = "\x{0394}" if $m eq "^";
+#		    $xt0 = $pr->text( $m, $xt0, $ychord, $fchord );
+#		}
+		$xt0 = $pr->text( $info->{adds}, $xt0,
+				   $ychord + $fchord->{size} * 0.2,
+				   $fchord,
+				   $fchord->{size} * 0.8,
+				 );
+		$xt0 = $pr->text( " ", $xt0, $ychord, $fchord );
+	    }
+	    else {
+		$xt0 = $pr->text( $chord." ", $x, $ychord, $fchord );
+	    }
 	    my $xt1 = $pr->text( $phrase, $x, $ytext, $ftext );
 	    $x = $xt0 > $xt1 ? $xt0 : $xt1;
 	}
@@ -793,86 +917,27 @@ sub songline {
     return;
 }
 
-# SMUFL mappings of common symbols.
-my %smufl =
-  ( brace		=> "\x{e000}",
-    reversedBrace	=> "\x{e001}",
-    barlineSingle	=> "\x{e030}",
-    barlineDouble	=> "\x{e031}",
-    barlineFinal	=> "\x{e032}",
-    repeatLeft		=> "\x{e040}",
-    repeatRight		=> "\x{e041}",
-    repeatRightLeft	=> "\x{e042}",
-    repeatDots		=> "\x{e043}",
-    dalSegno		=> "\x{e045}",
-    daCapo		=> "\x{e046}",
-    segno		=> "\x{e047}",
-    coda		=> "\x{e048}",
-    timeSig0		=> "\x{e080}", # timeSig1, ...etc...
-    flat		=> "\x{e260}",
-    sharp		=> "\x{e262}",
-    fermata		=> "\x{e4c0}",
-    repeat1Bar		=> "\x{e500}",
-    repeat2Bars		=> "\x{e501}",
-    repeat4Bars		=> "\x{e502}",
-    csymDiminished	=> "\x{e870}",
-    csymHalfDiminished	=> "\x{e871}",
-    csymAugmented	=> "\x{e872}",
-    csymMajorSeventh	=> "\x{e873}",
-    csymMinor		=> "\x{e874}",
-  );
-
-# Map ASCII bars (and pseudo-bar) to SMUFL code.
-my %sbmap =
-  ( "|"        => $smufl{barlineSingle},
-    "||"       => $smufl{barlineDouble},
-    "|."       => $smufl{barlineFinal},
-    "|:"       => $smufl{repeatLeft},
-    ":|"       => $smufl{repeatRight},
-    ":|:"      => $smufl{repeatRightLeft},
-    " %"       => $smufl{repeat2Bars},
-  );
-
-# Map ASCII and UTF8 measure repeats to SMUFL code.
-my %smap =
-  ( "%"        => $smufl{repeat1Bar},
-    "%%"       => $smufl{repeat2Bars},
-    "\x{2030}" => $smufl{repeat2Bars}, # permille
-  );
-
 sub is_bar {
-#    $_[0] =~ /^(\||\|\||\\|:|:\||\|\.)$/
-    $sbmap{$_[0]};
+    exists( $_[0]->{class} ) && $_[0]->{class} eq "bar";
 }
 
 sub gridline {
-    my ( $elt, $x, $y, $cellwidth, $barwidth, $ps ) = @_;
+    my ( $elt, $x, $y, $cellwidth, $barwidth, $margin, $ps ) = @_;
 
     # Grid context.
 
     my $pr = $ps->{pr};
     my $fonts = $ps->{fonts};
 
-    my $smufl = 1;		# use SMUFL font
-
-    $x += $barwidth/2;
+    $x += $barwidth;
+    $cellwidth += $barwidth;
 
+    # Use the chords font for the chords, and for the symbols size.
     my $fchord = { %{ $fonts->{grid} || $fonts->{chord} } };
     delete($fchord->{background});
-    my $schord;
-    if ( $smufl && $fonts->{symbols} ) {
-	$schord = { %{ $fonts->{symbols} } };
-    }
-    else {
-	$schord = { %{ $fchord } };
-	delete($schord->{background});
-	$smufl = 0;
-    }
-#    $schord->{size} = $fchord->{size};
-
     $y -= font_bl($fchord);
 
-    $elt->{tokens} //= [ '' ];
+    $elt->{tokens} //= [ {} ];
 
     my $firstbar;
     my $lastbar;
@@ -885,70 +950,190 @@ sub gridline {
     my $prevbar;
     my @tokens = @{ $elt->{tokens} };
     my $t;
+
+    if ( $margin->[0] ) {
+	$x -= $barwidth;
+	if ( $elt->{margin} ) {
+	    my $t = $elt->{margin};
+	    if ( $t->{chords} ) {
+		$t->{text} = "";
+		for ( 0..$#{ $t->{chords} } ) {
+		    $t->{text} .= $t->{chords}->[$_] . $t->{phrases}->[$_];
+		}
+	    }
+	    $pr->text( $t->{text}, $x, $y, $fonts->{comment} );
+	}
+	$x += $margin->[0] * $cellwidth + $barwidth;
+    }
+
     foreach my $i ( 0 .. $#tokens ) {
 	my $token = $tokens[$i];
-	if ( $t = is_bar($token) ) {
-	    $t = $token unless $smufl;
-	    $t = "{" if $t eq "|:";
-	    $t = "}" if $t eq ":|";
-	    $t = "}{" if $t eq ":|:";
-	    $pr->setfont($schord);
-	    my @t = ( $t );
-	    push( @t, $smufl{barlineSingle} ) if $t eq $smufl{repeat2Bars};
-	    for my $t ( @t ) {
-		my $w = $pr->strwidth($t);
-		my $y = $y;
-		$y += $schord->{size} / 2 if $t eq $smufl{repeat2Bars};
-		if ( defined $firstbar ) {
-		    my $x = $x;
-		    $x -= $w/2 if $i > $firstbar;
-		    $x -= $w/2 if $i == $lastbar;
-		    $pr->text( $t, $x, $y );
-		}
-		else {
-		    $pr->text( $t, $x + $barwidth/2 - $w/2, $y );
-		}
+	if ( exists $token->{chord} ) {
+	    $pr->text( $token->{chord}, $x, $y, $fchord )
+	      unless $token eq ".";
+	    $x += $cellwidth;
+	}
+	elsif ( $token->{class} eq "space" ) {
+	    $x += $cellwidth;
+	}
+	elsif ( $token->{class} eq "bar" ) {
+	    $x -= $barwidth;
+	    $t = $token->{symbol};
+	    if ( 0 ) {
+		$t = "{" if $t eq "|:";
+		$t = "}" if $t eq ":|";
+		$t = "}{" if $t eq ":|:";
+	    }
+	    else {
+		$t = "|:" if $t eq "{";
+		$t = ":|" if $t eq "}";
+		$t = ":|:" if $t eq "}{";
+	    }
+
+	    my $lcr = -1;	# left, center, right
+	    $lcr = 0 if $i > $firstbar;
+	    $lcr = 1 if $i == $lastbar;
+
+	    my $sz = $fchord->{size};
+
+	    if ( $t eq "|" ) {
+		pr_barline( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq "||" ) {
+		pr_dbarline( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq "|:" ) {
+		pr_rptstart( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq ":|" ) {
+		pr_rptend( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq ":|:" ) {
+		pr_rptendstart( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq "|." ) {
+		pr_endline( $x, $y, $lcr, $sz, $pr );
+	    }
+	    elsif ( $t eq " %" ) { # repeat2Bars
+		pr_repeat( $x+$sz/2, $y, 0, $sz, $pr );
+	    }
+	    else {
+		die($t);	# can't happen
 	    }
 	    $x += $barwidth;
 	    $prevbar = $i;
 	}
-	elsif ( ( $t = $smap{$token} || "" ) eq $smufl{repeat1Bar} ) {
-	    $t = $token unless $smufl;
+	elsif ( $token->{class} eq "repeat1" ) {
+	    $t = $token->{symbol};
 	    my $k = $prevbar + 1;
 	    while ( $k <= $#tokens
 		    && !is_bar($tokens[$k]) ) {
 		$k++;
 	    }
-	    $pr->setfont($schord);
-	    my $y = $y;
-	    $y += $schord->{size} / 2 if $t eq $smufl{repeat1Bar};
-	    my $w = $pr->strwidth($t);
-	    $pr->text( $t,
-		       $x + ($k - $prevbar - 1)*$cellwidth/2 - $w/2,
-		       $y );
+	    pr_repeat( $x + ($k - $prevbar - 1)*$cellwidth/2, $y,
+		       0, $fchord->{size}, $pr );
 	    $x += $cellwidth;
 	}
-	elsif ( ( $t = $smap{$token} || "" ) eq $smufl{repeat2Bars} ) {
+	elsif ( $token->{class} eq "repeat2" ) {
 	    # For repeat2Bars, change the next bar line to pseudo-bar.
 	    my $k = $prevbar + 1;
 	    while ( $k <= $#tokens
 		    && !is_bar($tokens[$k]) ) {
 		$k++;
 	    }
-	    $tokens[$k] = " %";
+	    $tokens[$k] = { symbol => " %", class => "bar" };
 	    $x += $cellwidth;
 	}
-	else {
-	    $pr->text( $token, $x, $y, $fchord )
-	      unless $token eq ".";
-	    $x += $cellwidth;
+	if ( $x > $ps->{papersize}->[0] ) {
+	    # This should be signalled by the parser.
+	    # warn("PDF: Too few cells for content\n");
+	    last;
 	}
     }
-    if ( $elt->{comment} ) {
-	$pr->text( " " . $elt->{comment}, $x, $y, $fonts->{comment} );
+
+    if ( $margin->[1] && $elt->{comment} ) {
+	my $t = $elt->{comment};
+	if ( $t->{chords} ) {
+	    $t->{text} = "";
+	    for ( 0..$#{ $t->{chords} } ) {
+		$t->{text} .= $t->{chords}->[$_] . $t->{phrases}->[$_];
+	    }
+	}
+	$pr->text( " " . $t->{text}, $x, $y, $fonts->{comment} );
     }
 }
 
+sub pr_barline {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = $w
+    $x -= $w / 2 * ($lcr + 1);
+    $pr->vline( $x, $y+0.9*$sz, $sz, $w );
+}
+
+sub pr_dbarline {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = 3 * $w
+    $x -= 1.5 * $w * ($lcr + 1);
+    $pr->vline( $x, $y+0.9*$sz, $sz, $w );
+    $x += 2 * $w;
+    $pr->vline( $x, $y+0.9*$sz, $sz, $w );
+}
+
+sub pr_rptstart {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = 3 * $w
+    $x -= 1.5 * $w * ($lcr + 1);
+    $pr->vline( $x, $y+0.9*$sz, $sz, $w );
+    $x += 2 * $w;
+    $y += 0.55 * $sz;
+    $pr->line( $x, $y, $x, $y+$w, $w );
+    $y -= 0.4 * $sz;
+    $pr->line( $x, $y, $x, $y+$w, $w );
+}
+
+sub pr_rptend {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = 3 * $w
+    $x -= 1.5 * $w * ($lcr + 1);
+    $pr->vline( $x + 2*$w, $y+0.9*$sz, $sz, $w );
+    $y += 0.55 * $sz;
+    $pr->line( $x, $y, $x, $y+$w, $w );
+    $y -= 0.4 * $sz;
+    $pr->line( $x, $y, $x, $y+$w, $w );
+}
+
+sub pr_rptendstart {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = 5 * $w
+    $x -= 2.5 * $w * ($lcr + 1);
+    $pr->vline( $x + 2*$w, $y+0.9*$sz, $sz, $w );
+    $y += 0.55 * $sz;
+    $pr->line( $x,      $y, $x     , $y+$w, $w );
+    $pr->line( $x+4*$w, $y, $x+4*$w, $y+$w, $w );
+    $y -= 0.4 * $sz;
+    $pr->line( $x,      $y, $x,      $y+$w, $w );
+    $pr->line( $x+4*$w, $y, $x+4*$w, $y+$w, $w );
+}
+
+sub pr_repeat {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 3;		# glyph width = 3 * $w
+    $x -= 1.5 * $w * ($lcr + 1);
+    my $lw = $sz / 10;
+    $x -= $w / 2;
+    $pr->line( $x, $y+0.2*$sz, $x + $w, $y+0.7*$sz, $lw );
+    $pr->line( $x, $y+0.6*$sz, $x + 0.07*$sz , $y+0.7*$sz, $lw );
+    $x += $w;
+    $pr->line( $x - 0.05*$sz, $y+0.2*$sz, $x + 0.02*$sz, $y+0.3*$sz, $lw );
+}
+
+sub pr_endline {
+    my ( $x, $y, $lcr, $sz, $pr ) = @_;
+    my $w = $sz / 10;		# glyph width = 2 * $w
+    $x -= 0.75 * $w * ($lcr + 1);
+    $pr->vline( $x, $y+0.85*$sz, 0.9*$sz, 2*$w );
+}
+
 sub imageline_vsp {
 }
 
@@ -969,15 +1154,13 @@ sub imageline {
 
     # Available width and height.
     my $pw;
-    if ( @{$ps->{columnoffsets}} > 1 ) {
+    if ( $ps->{columns} > 1 ) {
 	$pw = $ps->{columnoffsets}->[1]
 	  - $ps->{columnoffsets}->[0]
 	    - $ps->{columnspace};
     }
     else {
-	$pw = $ps->{papersize}->[0]
-	  - $ps->{marginleft}
-	    - $ps->{marginright};
+	$pw = $ps->{__rightmargin} - $ps->{_leftmargin};
     }
 
     my $ph = $ps->{papersize}->[1] - $ps->{margintop} - $ps->{marginbottom};
@@ -998,7 +1181,7 @@ sub imageline {
     }
     $h *= $scale;
     $w *= $scale;
-    $x += ($pw - $w) / 2 if $opts->{center} // 1;
+    $x += ($pw - $w) / 2 if $opts->{center};
 
     my $y = $gety->($h);	# may have been changed by checkspace
 
@@ -1018,13 +1201,14 @@ sub tocline {
     $y -= font_bl($ftoc);
     $pr->setfont($ftoc);
     $ps->{pr}->text( $elt->{title}, $x, $y );
-    my $p = $elt->{page} . ".";
+    my $p = $elt->{pageno} . ".";
     $ps->{pr}->text( $p, $x - 5 - $pr->strwidth($p), $y );
 
     my $ann = $pr->{pdfpage}->annotation;
-    $ann->link($pr->{pdf}->openpage($elt->{page}));
-    $ann->rect( $ps->{marginleft}, $y0 - $ftoc->{size},
-		$ps->{papersize}->[0]-$ps->{marginright}, $y0 );
+    $ann->link($elt->{page});
+    $ann->rect( $ps->{_leftmargin}, $y0 - $ftoc->{size},
+		$ps->{papersize}->[0]-$ps->{_rightmargin}, $y0 );
+    ####CHECK MARGIN RIGHT
 }
 
 sub has_visible_chords {
@@ -1054,6 +1238,9 @@ sub songline_vsp {
 
     return $vsp if $single_space && ! has_visible_chords($elt);
 
+    # No text printing if no text.
+    $vsp = 0 if join( "", @{ $elt->{phrases} } ) eq "";
+
     # We must show chords above lyrics, so add chords span.
     $vsp + $fonts->{chord}->{size} * $ps->{spacing}->{chords};
 }
@@ -1083,16 +1270,16 @@ sub text_vsp {
 
 sub chordgrid_vsp {
     my ( $elt, $ps ) = @_;
-    $ps->{fonts}->{chordgrid}->{size} * $ps->{spacing}->{chords}
-      + 0.40 * $ps->{chordgrid}->{width}
-	+ $ps->{chordgrid}->{vcells} * $ps->{chordgrid}->{height}
-	  + $ps->{chordgrid}->{vspace} * $ps->{chordgrid}->{height};
+    $ps->{fonts}->{diagram}->{size} * $ps->{spacing}->{chords}
+      + 0.40 * $ps->{diagrams}->{width}
+	+ $ps->{diagrams}->{vcells} * $ps->{diagrams}->{height}
+	  + $ps->{diagrams}->{vspace} * $ps->{diagrams}->{height};
 }
 
 sub chordgrid_hsp {
     my ( $elt, $ps ) = @_;
-    App::Music::ChordPro::Chords::strings() * $ps->{chordgrid}->{width}
-      + $ps->{chordgrid}->{hspace} * $ps->{chordgrid}->{width};
+    App::Music::ChordPro::Chords::strings() * $ps->{diagrams}->{width}
+      + $ps->{diagrams}->{hspace} * $ps->{diagrams}->{width};
 }
 
 my @Roman = qw( I II III IV V VI VI VII VIII IX X XI XII );
@@ -1101,14 +1288,23 @@ sub chordgrid {
     my ( $name, $x, $y, $ps ) = @_;
     my $x0 = $x;
 
-    my $gw = $ps->{chordgrid}->{width};
-    my $gh = $ps->{chordgrid}->{height};
+    my $gw = $ps->{diagrams}->{width};
+    my $gh = $ps->{diagrams}->{height};
     my $dot = 0.80 * $gw;
-    my $lw  = 0.10 * $gw;
+    my $lw  = ($ps->{diagrams}->{linewidth} || 0.10) * $gw;
 
-    my $info = App::Music::ChordPro::Chords::chord_info($name);
+    my $info;
+    if ( eval{ $name->{name} } ) {
+	$info = $name;
+	$info->{origin} = 0;
+	$name = $info->{name};
+	$info->{base}--;
+    }
+    else {
+	$info = App::Music::ChordPro::Chords::chord_info($name);
+    }
     unless ( $info ) {
-	warn("PDF: Skipping grid for unknown chord $name\n");
+	warn("PDF: Skipping diagram for unknown chord $name\n");
 	return;
     }
 
@@ -1118,44 +1314,59 @@ sub chordgrid {
     my $w = $gw * ($strings - 1);
 
     # Draw font name.
-    my $font = $ps->{fonts}->{chordgrid};
+    my $font = $ps->{fonts}->{diagram};
     $pr->setfont($font);
     $name .= "*"
-      unless $info->{builtin} || $::config->{chordgrid}->{show} eq "user";
+      unless $info->{origin} <= 1 || $::config->{diagrams}->{show} eq "user";
     $pr->text( $name, $x + ($w - $pr->strwidth($name))/2, $y - font_bl($font) );
-    $y -= $font->{size} * $ps->{spacing}->{chords} + $dot/2;
+    $y -= $font->{size} * $ps->{spacing}->{chords} + $dot/2 + $lw;
 
     if ( $info->{base} > 0 ) {
 	my $i = @Roman[$info->{base}] . "  ";
-	$pr->setfont( $ps->{fonts}->{chordgrid_capo}, $gh );
+	$pr->setfont( $ps->{fonts}->{diagram_base}, $gh );
 	$pr->text( $i, $x-$pr->strwidth($i), $y-$gh/2,
-		   $ps->{fonts}->{chordgrid_capo}, $gh );
+		   $ps->{fonts}->{diagram_base}, $gh );
     }
 
-    my $v = $ps->{chordgrid}->{vcells};
+    my $v = $ps->{diagrams}->{vcells};
     my $h = $strings;
     $pr->hline( $x, $y - $_*$gh, $w, $lw ) for 0..$v;
-    $pr->vline( $x + $_*$gw, $y, $gh*$v, $lw ) for 0..$h-1;
 
     $x -= $gw/2;
-    for my $fret ( @{ $info->{strings} } ) {
+    for my $sx ( 0 .. @{ $info->{strings} }-1 ) {
+	my $fret = $info->{strings}->[$sx];
+	my $fing;
+	$fing = $info->{fingers}->[$sx] if $info->{fingers};
 	if ( $fret > 0 ) {
-	    $pr->circle( $x+$gw/2, $y-$fret*$gh+$gh/2, $dot/2, $lw,
-			 "black", "black");
+	    my $glyph = "\x{6c}";
+	    if ( $fing && $fing > 0 ) {
+		# Note: circle must go under the text.
+		$pr->circle( $x+$gw/2, $y-$fret*$gh+$gh/2, $dot/2, $lw,
+			     "white", undef);
+		$glyph = pack( "C", 0xca + $fing - 1 );
+	    }
+	    my $dot = $dot/0.7;
+	    $pr->setfont( $ps->{fonts}->{chordfingers}, $dot );
+	    $pr->text( $glyph,
+		       $x+$gw/2-$pr->strwidth($glyph)/2,
+		       $y-$fret*$gh+$gh/2-$pr->strwidth($glyph)/2+$lw/2,
+		       $ps->{fonts}->{chordfingers}, $dot ) ;
 	}
 	elsif ( $fret < 0 ) {
-	    $pr->cross( $x+$gw/2, $y+$gh/3, $dot/3, $lw, "black");
+	    $pr->cross( $x+$gw/2, $y+$lw+$gh/3, $dot/3, $lw, "black");
 	}
 	elsif ( $info->{base} >= 0 ) {
-	    $pr->circle( $x+$gw/2, $y+$gh/3, $dot/3, $lw,
+	    $pr->circle( $x+$gw/2, $y+$lw+$gh/3, $dot/3, $lw,
 			 undef, "black");
 	}
     }
     continue {
 	$x += $gw;
     }
+    # Note: vline must go under the circle.
+    $pr->vline( $x0 + $_*$gw, $y, $gh*$v, $lw ) for 0..$h-1;
 
-    return $gw * ( $ps->{chordgrid}->{hspace} + $strings );
+    return $gw * ( $ps->{diagrams}->{hspace} + $strings );
 }
 
 sub set_columns {
@@ -1166,16 +1377,22 @@ sub set_columns {
     else {
 	$ps->{columns} = $cols ||= 1;
     }
-    $ps->{columnoffsets} = [ 0 ];
-    return unless $cols > 1;
 
+    # Note that _leftmargin and _rightmargin may not yet have been set.
+    # Luckily, marginleft + marginright is always equal to
+    # _leftmargin + _rightmargin.
     my $w = $ps->{papersize}->[0]
-      - $ps->{marginright} - $ps->{marginleft};
+      - $ps->{marginleft} - $ps->{marginright};
+
+    $ps->{columnoffsets} = [ 0 ];
+     push( @{ $ps->{columnoffsets} }, $w ), return unless $cols > 1;
+
     my $d = ( $w - ( $cols - 1 ) * $ps->{columnspace} ) / $cols;
     $d += $ps->{columnspace};
     for ( 1 .. $cols-1 ) {
 	push( @{ $ps->{columnoffsets} }, $_ * $d );
     }
+    push( @{ $ps->{columnoffsets} }, $w );
 }
 
 sub showlayout {
@@ -1184,34 +1401,35 @@ sub showlayout {
     my $col = "black";
     my $lw = 0.5;
 
-    $pr->rectxy( $ps->{marginleft},
+    my $mr = $ps->{_rightmargin};
+    my $ml = $ps->{_leftmargin};
+
+    $pr->rectxy( $ml,
 		 $ps->{marginbottom},
-		 $ps->{papersize}->[0]-$ps->{marginright},
+		 $ps->{papersize}->[0]-$mr,
 		 $ps->{papersize}->[1]-$ps->{margintop},
 		 $lw, undef, $col);
 
-    $pr->hline( $ps->{marginleft},
-		$ps->{papersize}->[1]-$ps->{margintop}+$ps->{headspace},
-		$ps->{papersize}->[0]-$ps->{marginleft}-$ps->{marginright},
-		$lw, $col );
-
-    $pr->hline( $ps->{marginleft},
-		$ps->{marginbottom}-$ps->{footspace},
-		$ps->{papersize}->[0]-$ps->{marginleft}-$ps->{marginright},
-		$lw, $col );
+    my @a = ( $ml,
+	      $ps->{papersize}->[1]-$ps->{margintop}+$ps->{headspace},
+	      $ps->{papersize}->[0]-$ml-$mr,
+	      $lw, $col );
+    $pr->hline(@a);
+    $a[1] = $ps->{marginbottom}-$ps->{footspace};
+    $pr->hline(@a);
 
     my @off = @{ $ps->{columnoffsets} };
+    pop(@off);
     @off = ( $ps->{chordscolumn} ) if $ps->{chordscolumn};
     foreach my $i ( 0 .. @off-1 ) {
 	next unless $off[$i];
-	$pr->vline( $ps->{marginleft}+$off[$i],
-		    $ps->{marginbottom},
-		    $ps->{margintop}-$ps->{papersize}->[1]+$ps->{marginbottom},
-		    $lw, $col );
-	$pr->vline( $ps->{marginleft}+$off[$i]-$ps->{columnspace},
-		    $ps->{marginbottom},
-		    $ps->{margintop}-$ps->{papersize}->[1]+$ps->{marginbottom},
-		    $lw, $col );
+	@a = ( $ml+$off[$i],
+	       $ps->{marginbottom},
+	       $ps->{margintop}-$ps->{papersize}->[1]+$ps->{marginbottom},
+	       $lw, $col );
+	$pr->vline(@a);
+	$a[0] -= $ps->{columnspace};
+	$pr->vline(@a);
     }
 }
 
@@ -1265,31 +1483,28 @@ sub configurator {
 
     # Chord grid width.
     if ( $options->{'chord-grid-size'} ) {
-	$pdf->{chordgrid}->{width} =
-	  $pdf->{chordgrid}->{height} =
+	$pdf->{diagrams}->{width} =
+	  $pdf->{diagrams}->{height} =
 	    $options->{'chord-grid-size'} /
 	      App::Music::ChordPro::Chords::strings();
     }
 
     # Add font dirs.
-    my $fontdir = $pdf->{fontdir} || $ENV{FONTDIR};
-    if ( $fontdir ) {
+    for my $fontdir ( $pdf->{fontdir}, ::findlib("fonts"), $ENV{FONTDIR} ) {
+	next unless $fontdir;
 	if ( -d $fontdir ) {
-	    PDF::API2::addFontDirs($fontdir);
+	    $pdfapi->can("addFontDirs")->($fontdir);
 	}
 	else {
 	    warn("PDF: Ignoring fontdir $fontdir [$!]\n");
 	    undef $fontdir;
 	}
     }
-    else {
-	undef $fontdir;
-    }
 
     # Map papersize name to [ width, height ].
     unless ( eval { $pdf->{papersize}->[0] } ) {
-	require PDF::API2::Resource::PaperSizes;
-	my %ps = PDF::API2::Resource::PaperSizes->get_paper_sizes;
+	eval "require ${pdfapi}::Resource::PaperSizes";
+	my %ps = "${pdfapi}::Resource::PaperSizes"->get_paper_sizes;
 	die("Unhandled paper size: ", $pdf->{papersize}, "\n")
 	  unless exists $ps{lc $pdf->{papersize}};
 	$pdf->{papersize} = $ps{lc $pdf->{papersize}}
@@ -1302,9 +1517,11 @@ sub configurator {
     $fonts->{comment}        ||= { %{ $fonts->{text}  } };
     $fonts->{toc}	     ||= { %{ $fonts->{text}  } };
     $fonts->{empty}	     ||= { %{ $fonts->{text}  } };
-    $fonts->{grid}           ||= { %{ $fonts->{comment} } };
-    $fonts->{chordgrid}      ||= { %{ $fonts->{comment} } };
-    $fonts->{chordgrid_capo} ||= { %{ $fonts->{text} } };
+    $fonts->{grid}           ||= { %{ $fonts->{chord} } };
+    $fonts->{grid_margin}    ||= { %{ $fonts->{comment} } };
+    $fonts->{diagram}        ||= { %{ $fonts->{comment} } };
+    $fonts->{diagram_base}   ||= { %{ $fonts->{comment} } };
+    $fonts->{chordfingers}     = { name => 'ZapfDingbats' };
     $fonts->{subtitle}->{size}       ||= $fonts->{text}->{size};
     $fonts->{comment_italic}->{size} ||= $fonts->{text}->{size};
     $fonts->{comment_box}->{size}    ||= $fonts->{text}->{size};
@@ -1338,20 +1555,22 @@ sub fmt_subst {
 # Three-part titles.
 # Note: baseline printing.
 sub tpt {
-    my ( $ps, $class, $type, $leftpage, $x, $y, $s ) = @_;
+    my ( $ps, $class, $type, $rightpage, $x, $y, $s ) = @_;
     my $fmt = get_format( $ps, $class, $type );
     return unless $fmt;
 
     # @fmt = ( left-fmt, center-fmt, right-fmt )
-
-    my @fmt = @{$fmt};
-    @fmt = @fmt[2,1,0] if $leftpage; # swap
+    unless ( @$fmt == 3 ) {
+	die("ASSERT: " . scalar(@$fmt)," part format $class $type");
+    }
+    my @fmt = ( @$fmt );
+    @fmt = @fmt[2,1,0] unless $rightpage; # swap
 
     my $pr = $ps->{pr};
     my $font = $ps->{fonts}->{$type};
 
     $pr->setfont($font);
-    my $rm = $ps->{papersize}->[0] - $ps->{marginright};
+    my $rm = $ps->{papersize}->[0] - $ps->{_rightmargin};
 
     # Left part. Easiest.
     $pr->text( fmt_subst( $s, $fmt[0] ), $x, $y ) if $fmt[0];
@@ -1378,7 +1597,6 @@ package PDFWriter;
 
 use strict;
 use warnings;
-use PDF::API2;
 use Encode;
 
 my $faketime = 1465041600;
@@ -1386,9 +1604,9 @@ my $faketime = 1465041600;
 my %fontcache;			# speeds up 2 seconds per song
 
 sub new {
-    my ( $pkg, $ps, @file ) = @_;
+    my ( $pkg, $ps ) = @_;
     my $self = bless { ps => $ps }, $pkg;
-    $self->{pdf} = PDF::API2->new( -file => $file[0] );
+    $self->{pdf} = $pdfapi->new;
     $self->{pdf}->{forcecompress} = 0 if $regtest;
     $self->{pdf}->mediabox( $ps->{papersize}->[0],
 			    $ps->{papersize}->[1] );
@@ -1415,7 +1633,6 @@ sub text {
 
     $self->setfont($font, $size);
 
-    $text = encode( "cp1250", $text ) unless $font->{file};
     if ( $font->{color} ) {
 	$self->{pdftext}->strokecolor( $font->{color} );
 	$self->{pdftext}->fillcolor( $font->{color} );
@@ -1448,11 +1665,25 @@ sub strwidth {
     $self->{pdftext}->advancewidth($text);
 }
 
+sub line {
+    my ( $self, $x0, $y0, $x1, $y1, $lw, $color ) = @_;
+    my $gfx = $self->{pdfpage}->gfx;
+    $gfx->save;
+    $gfx->strokecolor($color ||= "black");
+    $gfx->linecap(1);
+    $gfx->linewidth($lw||1);
+    $gfx->move( $x0, $y0 );
+    $gfx->line( $x1, $y1 );
+    $gfx->stroke;
+    $gfx->restore;
+}
+
 sub hline {
     my ( $self, $x, $y, $w, $lw, $color ) = @_;
     my $gfx = $self->{pdfpage}->gfx;
     $gfx->save;
     $gfx->strokecolor($color ||= "black");
+    $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->move( $x, $y );
     $gfx->hline( $x + $w );
@@ -1462,9 +1693,10 @@ sub hline {
 
 sub vline {
     my ( $self, $x, $y, $h, $lw, $color ) = @_;
-    my $gfx = $self->{pdfpage}->gfx;
+    my $gfx = $self->{pdfpage}->gfx(1); # under
     $gfx->save;
     $gfx->strokecolor($color ||= "black");
+    $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->move( $x, $y );
     $gfx->vline( $y - $h );
@@ -1478,6 +1710,7 @@ sub rectxy {
     $gfx->save;
     $gfx->strokecolor($strokecolor) if $strokecolor;
     $gfx->fillcolor($fillcolor) if $fillcolor;
+    $gfx->linecap(2);
     $gfx->linewidth($lw||1);
     $gfx->rectxy( $x, $y, $x1, $y1 );
     $gfx->close;
@@ -1488,7 +1721,7 @@ sub rectxy {
 
 sub circle {
     my ( $self, $x, $y, $r, $lw, $fillcolor, $strokecolor ) = @_;
-    my $gfx = $self->{pdfpage}->gfx;
+    my $gfx = $self->{pdfpage}->gfx(1); # under
     $gfx->save;
     $gfx->strokecolor($strokecolor) if $strokecolor;
     $gfx->fillcolor($fillcolor) if $fillcolor;
@@ -1542,9 +1775,9 @@ sub add_image {
 }
 
 sub newpage {
-    my ( $self, $ps ) = @_;
+    my ( $self, $ps, $page ) = @_;
     #$self->{pdftext}->textend if $self->{pdftext};
-    $self->{pdfpage} = $self->{pdf}->page;
+    $self->{pdfpage} = $self->{pdf}->page($page);
     $self->{pdfpage}->mediabox( $ps->{papersize}->[0],
 				$ps->{papersize}->[1] );
     $self->{pdftext} = $self->{pdfpage}->text;
@@ -1557,8 +1790,16 @@ sub add {
 }
 
 sub finish {
-    my $self = shift;
-    $self->{pdf}->save;
+    my ( $self, $file ) = @_;
+
+    if ( $file && $file ne "-" ) {
+	$self->{pdf}->saveas($file);
+    }
+    else {
+	binmode(STDOUT);
+	print STDOUT ( $self->{pdf}->stringify );
+	close(STDOUT);
+    }
 }
 
 sub init_fonts {
@@ -1601,7 +1842,7 @@ sub init_font {
     else {
 	$font->{font} =
 	  $fontcache{"__core__".$font->{name}} ||=
-	  $self->{pdf}->corefont( $font->{name}, -dokern => 1 );
+	    $self->{pdf}->corefont( $font->{name}, -dokern => 1 );
     }
 
     unless ( $font->{font} ) {
